# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

# A tuple containing the five components of the version number: major, minor,
# micro, releaselevel, and serial. All values except releaselevel are integers;
# the release level is 'dev' or 'final'. The
# version_info value corresponding to the byov version 2.0 is (2, 0, 0,
# 'final', 0).
import os

# Using realpath below as __file__ may include a symlink
HERE = os.path.realpath(os.path.dirname(__file__))
BRANCH_ROOT = os.path.abspath(os.path.join(HERE, '..'))

__version__ = (2, 3, 9, 'final', 0)


# FIXME: This is not exposed to user like config options. doc ?
# -- vila 2022-03-31

# A list of directories where scripts and hooks (as relative paths) are
# searched for. '~' is expanded if present. An exception is raised if the
# relative path cannot be found in any of the directories or if it's not an
# executable.

# This can be extended by other libraries/projects to layout a new set of
# configuration files on top of the regular ones.
path = os.getenv('BYOV_PATH', os.getcwd()).split(':')


def version(ver=None):
    if ver is None:
        ver = __version__
    major, minor, patch, ver_type, increment = ver
    if ver_type == 'final':
        return '{}.{}.{}'.format(major, minor, patch)
    else:
        return '{}.{}.{}{}{}'.format(*ver)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import unicode_literals

import errno
import unittest

from byov import (
    config,
    errors,
)
from byov.tests import (
    features as features,
    fixtures as fixtures,
)
from byov.vms import lxd


class TestPull(unittest.TestCase):

    def setUp(self):
        super(TestPull, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-debian')
        fixtures.set_uniq_vm_name(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'ephemeral-lxd')
        conf.set('vm.backing', 'byov-tests-lxd-debian')
        conf.store.save()
        conf.store.unload()
        self.vm = lxd.EphemeralLxd(config.VmStack(self.vm_name))
        self.addCleanup(self.vm.stop)
        self.vm.start()

    def test_simple(self):
        remote = '/home/{}/bar'.format(self.vm.conf.get('vm.user'))
        ret, _, _ = self.vm.shell_captured('echo', 'quux', '>' + remote)
        self.assertEqual(0, ret)
        local = 'foo'
        ret = self.vm.pull(remote, local)
        self.assertEqual(0, ret)
        ret, _, _ = self.vm.shell_captured('test', '-f', remote)
        self.assertEqual(0, ret)

    def test_unknown_src(self):
        remote = '/home/{}/bar'.format(self.vm.conf.get('vm.user'))
        local = 'foo'
        with self.assertRaises(errors.CommandError):
            self.vm.pull(remote, local)

    def test_unknown_dst(self):
        remote = '/home/{}/bar'.format(self.vm.conf.get('vm.user'))
        ret, _, _ = self.vm.shell_captured('echo', 'quux', '>' + remote)
        self.assertEqual(0, ret)
        local = 'I-dont-exist/foo'
        with self.assertRaises(IOError) as cm:
            self.vm.pull(remote, local)
        self.assertEqual(2, cm.exception.args[0])
        # aka FileNotFound (python2 specific)
        self.assertEqual(errno.ENOENT, cm.exception.errno)
        self.assertEqual(local, cm.exception.filename)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import errno
import logging
import os
import subprocess


from byov import errors


logger = logging.getLogger(__name__)


def run(args, cmd_input=None, raise_on_error=True):
    """Run the specified command capturing output and errors.

    :param args: A list of a command and its arguments.

    :param cmd_input: A unicode string to feed the command with.

    :param raise_on_error: A boolean controlling whether or not an exception is
        raised if the command fails.

    :return: A tuple of the return code, the output and the errors as unicode
        strings.

    """
    stdin = None
    if cmd_input is not None:
        stdin = subprocess.PIPE
        cmd_input = cmd_input.encode('utf8')
    logger.debug('Running {}'.format(' '.join(args)))
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            stdin=stdin)
    out, err = proc.communicate(cmd_input)
    out = out.decode('utf8')
    err = err.decode('utf8')
    if raise_on_error and proc.returncode:
        # FIXME: cmd_input should be exposed if not None -- vila 2022-05-12
        raise errors.CommandError(args, proc.returncode, out, err)
    logger.debug('Returned: {}'.format(proc.returncode))
    if out:
        logger.debug('stdout: {}'.format(out))
    if err:
        logger.debug('stderr: {}'.format(err))
    return proc.returncode, out, err


def raw_run(args):
    """Run the specified command without redirecting anything.

    This is mainly useful to wrap a command leaving std{in,out,err}
    untouched so the user can interact.

    :param args:  A list of a command and its arguments.

    :return: The return code.
    """
    logger.debug('Running {}'.format(' '.join(args)))
    proc = subprocess.Popen(args)
    proc.wait()
    return proc.returncode


def pipe(args):
    """Run the specified command as a pipe.

    The caller is responsible for processing the output and the errors.

    :param args:  A list of a command and its arguments.

    :return: The Popen object.
    """
    proc = subprocess.Popen(args,
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return proc


def which(cmd, path, mode=os.F_OK | os.X_OK):
    """Given a command, mode, and a PATH list of strings, return the path which
    conforms to the given mode on the PATH, or None if there is no such file.

    `mode` defaults to os.F_OK | os.X_OK.

    This is heavily inspired by shutils.which but the API and implementation is
    significantly different and the implementation way simpler.
    """
    if path:
        files = [cmd]

        seen = set()
        for d in path:
            normdir = os.path.normcase(d)
            if normdir not in seen:
                seen.add(normdir)
                for thefile in files:
                    name = os.path.join(d, os.path.normpath(thefile))
                    if os.path.exists(name):
                        if os.access(name, mode):
                            return name
                        else:
                            raise PermissionError(
                                errno.EACCES,
                                name + ' is not executable', cmd)
        raise FileNotFoundError(errno.ENOENT, '{} not found in: {}'.format(
            cmd, path), cmd)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018-2023 Vincent Ladeuil.
# Copyright 2014-2017 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import errno
import os
import unittest

from byot import (
    fixtures,
    tests,
)
from byov import (
    errors,
    subprocesses,
)


class TestRun(unittest.TestCase):

    def setUp(self):
        super(TestRun, self).setUp()
        fixtures.set_uniq_cwd(self)

    def assertFileNotFound(self, exception, file_name):
        self.assertEqual(errno.ENOENT, exception.errno)
        if tests.python.version > (3, 7):
            self.assertEqual("No such file or directory",
                             exception.strerror)
        else:
            self.assertEqual(
                "No such file or directory: '{}'".format(file_name),
                exception.strerror)
        if tests.python.version > (3, 6):
            self.assertEqual(file_name, exception.filename)
        else:
            self.assertEqual(None, exception.filename)

    def test_success(self):
        ret, out, err = subprocesses.run(['echo', '-n', 'Hello'])
        self.assertEqual(0, ret)
        self.assertEqual('Hello', out)
        self.assertEqual('', err)

    def test_with_input(self):
        ret, out, err = subprocesses.run(['cat', '-'], 'bar\n')
        self.assertEqual(0, ret)
        self.assertEqual('bar\n', out)
        self.assertEqual('', err)

    def test_failure(self):
        with self.assertRaises(errors.CommandError) as cm:
            subprocesses.run(['ls', 'I-dont-exist'])
        self.assertEqual(2, cm.exception.retcode)
        self.assertEqual('', cm.exception.out)
        self.assertTrue('I-dont-exist' in cm.exception.err)

    def test_raise_on_error(self):
        retcode, out, err = subprocesses.run(
            ['ls', 'I-dont-exist'], raise_on_error=False)
        self.assertEqual(2, retcode)
        self.assertEqual('', out)
        self.assertTrue('I-dont-exist' in err)

    def test_error(self):
        with self.assertRaises(FileNotFoundError) as cm:
            subprocesses.run(['I-dont-exist'])
        self.assertFileNotFound(cm.exception, 'I-dont-exist')


class TestPipe(TestRun):

    def test_success(self):
        proc = subprocesses.pipe(['echo', '-n', 'Hello'])
        self.assertEqual('Hello', proc.stdout.read().decode('utf8'))

    def test_failure(self):
        proc = subprocesses.pipe(['ls', 'I-dont-exist'])
        self.assertTrue('I-dont-exist' in proc.stdout.read().decode('utf8'))
        # We get the above in stdout because stderr is redirected there
        self.assertIs(None, proc.stderr)

    def test_error(self):
        with self.assertRaises(FileNotFoundError) as cm:
            subprocesses.pipe(['I-dont-exist'])
        self.assertFileNotFound(cm.exception, 'I-dont-exist')


class TestWhich(unittest.TestCase):

    def setUp(self):
        super().setUp()
        fixtures.set_uniq_cwd(self)

    def test_empty_path(self):
        self.assertEqual(None, subprocesses.which('/bin/true', []))

    def test_current_dir(self):
        with open('script', 'w') as f:
            f.write('''#!/bin/sh\ntrue\n''')
        os.chmod('script', 0o755)
        self.assertEqual('./script', subprocesses.which('script', ['.']))

    def test_subdir(self):
        os.mkdir('sub')
        path = os.path.join('sub', 'script')
        with open(path, 'w') as f:
            f.write('''#!/bin/sh\ntrue\n''')
        os.chmod(path, 0o755)
        self.assertEqual('sub/script', subprocesses.which(
            'script', ['.', 'sub']))

# MISSINGTESTS:
# test_ssh_no_args (hard, interactive session)
# test_ssh_simple_command
# test_ssh_cant_connect (wrong host, unknown host, missing key, wrong user)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2017-2018 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import errno
import io
import os
import unittest

import byov
from byov import (
    config,
    errors,
)
from byov.tests import (
    features,
    fixtures,
)
from byov.vms import lxd


class TestSetupHook(unittest.TestCase):

    def setUp(self):
        super(TestSetupHook, self).setUp()
        fixtures.setup_tests_config(self)
        features.requires_existing_vm(self, 'byov-tests-lxd-ubuntu')
        fixtures.set_uniq_vm_name(self)
        self.out = io.StringIO()
        self.err = io.StringIO()
        fixtures.override_logging(self)
        # Create a shared config
        conf = config.VmStack(self.vm_name)
        conf.set('vm.name', self.vm_name)
        conf.set('vm.class', 'ephemeral-lxd')
        conf.set('vm.backing', 'byov-tests-lxd-ubuntu')
        conf.store.save()
        conf.store.unload()

    def test_empty_hook(self):
        conf = config.VmStack(self.vm_name)
        conf.set('vm.setup.hook', '')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        # Just a smoke test
        self.vm.start()

    def test_non_existing_hook(self):
        conf = config.VmStack(self.vm_name)
        conf.set('vm.setup.hook', 'i-dont-exist')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.start()
        self.assertEqual(127, cm.exception.retcode)
        self.assertEqual('/bin/sh: 1: i-dont-exist: not found\n',
                         cm.exception.err)

    def test_hook_success(self):
        conf = config.VmStack(self.vm_name)
        conf.set('vm.setup.hook', 'ls -d .')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        # Just a smoke test
        self.vm.start()

    def test_hook_script_success(self):
        fixtures.patch(self, byov, 'path', byov.path + [self.uniq_dir])
        conf = config.VmStack(self.vm_name)
        conf.set('content', 'true')
        with open('script', 'w') as f:
            f.write('''#!/bin/sh\n{content}\n''')
        os.chmod('script', 0o755)
        conf.set('vm.setup.hook', '@script')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        # Just a smoke test
        self.vm.start()

    def test_hook_script_fails(self):
        fixtures.patch(self, byov, 'path', byov.path + [self.uniq_dir])
        conf = config.VmStack(self.vm_name)
        conf.set('content', 'false')
        with open('script', 'w') as f:
            f.write('''#!/bin/sh -e\n{content}\ntrue\n''')
        os.chmod('script', 0o755)
        conf.set('vm.setup.hook', '@script')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        with self.assertRaises(errors.CommandError) as cm:
            self.vm.start()
        self.assertEqual(1, cm.exception.retcode)

    def test_hook_non_executable_script_fails(self):
        fixtures.patch(self, byov, 'path', byov.path + [self.uniq_dir])
        conf = config.VmStack(self.vm_name)
        conf.set('content', 'false')
        with open('script', 'w') as f:
            f.write('''#!/bin/sh -e\n{content}\ntrue\n''')
        conf.set('vm.setup.hook', '@script')
        self.vm = lxd.EphemeralLxd(conf)
        self.addCleanup(self.vm.stop)
        with self.assertRaises(PermissionError) as cm:
            self.vm.start()
        self.assertEqual(errno.EACCES, cm.exception.errno)



# This file is part of Build Your Own Virtual machine.
#
# Copyright 2018-2023 Vincent Ladeuil.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from byov import vms


logger = logging.getLogger(__name__)


class Ssh(vms.VM):
    """A minimal ssh vm.

    This is mainly to allow connection and scripts remote execution.
    """

    def state(self):
        return 'RUNNING'

    def setup(self):
        raise NotImplementedError

    def start(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError
